

import os
import typing
import tarfile

import jk_typing
import jk_utils
import jk_logging
import jk_json
import jk_prettyprintobj





class PythonNativeTar(object):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	@jk_typing.checkFunctionSignature()
	def __init__(self):
		pass
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	#
	# Tar the contents of a directory.
	# An exception is raised on error.
	# Before an error is raised the logger may be used to provide a more detailed error message.
	#
	# @param	str absDestTarFilePath				The (absolute) output file path
	# @param	str absSrcDirPath					The (absolute) source directory that contains all files and subdirectories to include
	# @param	str[] filesAndDirsToInclude			The files and directory names to include
	# @param	AbstractLogger log					The logger to use (if there is anything to log, e.g. error information)
	# @param	Callable[TarInfo]:TarInfo filter	(optional) An optional filter function that is directly passed on to the
	#												python tarfile API. This filter function receives a TarInfo record of
	#												the current file to pack and can either return the same or a modified
	#												version of the TarInfo record or None to indicate that
	#												this entry should be excluded.
	#
	@jk_typing.checkFunctionSignature()
	def tarDirContents(
			self,
			absDestTarFilePath:str,
			absSrcDirPath:str,
			filesAndDirsToInclude:typing.List[str],
			log:jk_logging.AbstractLogger,
			*,
			filter:typing.Callable[[tarfile.TarInfo],typing.Union[tarfile.TarInfo,None]] = None,
		) -> None:

		with tarfile.open(absDestTarFilePath, "w") as tarOut:
			for fileName in filesAndDirsToInclude:
				srcFilePath = os.path.join(absSrcDirPath, fileName)
				tarOut.add(srcFilePath, fileName, recursive=True, filter=filter)
	#

	def listTarContents(
			self,
			absTarFilePath:str,
			log:jk_logging.AbstractLogger,
		) -> typing.List[str]:

		ret = []
		with tarfile.open(absTarFilePath, "r") as tarIn:
			for tarinfo in tarIn:
				ret.append(tarinfo.name)

		return ret
	#

	@jk_typing.checkFunctionSignature()
	def untarToDir(self,
			absTarFilePath:str,
			absDestDirPath:str,
			log:jk_logging.AbstractLogger,
		) -> None:

		if not os.path.isdir(absDestDirPath):
			os.makedirs(absDestDirPath)

		with tarfile.open(absTarFilePath, "r") as tarIn:
			tarIn.extractall(absDestDirPath)
	#

#












__api_version__ = 2
__author__ = "Jürgen Knauth"
__version__ = "0.2022.10.25"



from .FileUncompressionGuess import FileUncompressionGuess

from .SpoolInfo import SpoolInfo
from .Spooler import Spooler

from .Packer import Packer
from .Packer1 import Packer1
from .Unpacker import Unpacker
from .Unpacker1 import Unpacker1






import os
import typing
import re

import jk_typing
import jk_logging
import jk_json
import jk_prettyprintobj





class _NameMatcher(object):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	# @param		(str,Pattern)[] namePatterns			(optional) A list of names or name patterns
	# @param		bool bDefaultIfNone						(optional) What to return if no named pattern has been given: True or False
	#
	def __init__(self, namePatterns:typing.Iterable[typing.Union[str,re.Pattern]] = None, bDefaultIfNone:bool = True) -> None:
		if namePatterns is None:
			self.__bReturnImmediately = True
			self.__names = None
			self.__namePatterns = None

		else:
			self.__bReturnImmediately = False
			self.__names = []
			self.__namePatterns = []
			for x in namePatterns:
				if isinstance(x, str):
					self.__names.append(x)
				elif isinstance(x, re.Pattern):
					self.__namePatterns.append(x)
				else:
					raise TypeError(repr(type(x)))
			self.__names = set(self.__names)

		self.__bDefaultIfNone = bDefaultIfNone
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def match(self, name:str) -> bool:
		assert isinstance(name, str)

		if self.__bReturnImmediately:
			return self.__bDefaultIfNone

		if name in self.__names:
			return True

		for x in self.__namePatterns:
			assert isinstance(x, re.Pattern)
			m = x.match(name)
			if m:
				return True

		return False
	#

	@jk_typing.checkFunctionSignature()
	def matchDebug(self, name:str, posActivity:str, negActivity:str, log:jk_logging.AbstractLogger) -> bool:
		if self.__bReturnImmediately:
			log.debug("{}, always True -> {}".format(repr(name), posActivity if self.__bDefaultIfNone else negActivity))
			return self.__bDefaultIfNone

		if name in self.__names:
			log.debug("{}, in {} -> {}".format(repr(name), self.__names, posActivity))
			return True

		for x in self.__namePatterns:
			assert isinstance(x, re.Pattern)
			m = x.match(name)
			if m:
				log.debug("{}, matches pattern {} -> {}".format(repr(name), x, posActivity))
				return True

		log.debug("{}, does not match -> {}".format(repr(name), negActivity))
		return False
	#

#









import os
import typing
import tarfile
import gzip
import bz2
import lzma
import zipfile

import jk_simpleexec
import jk_logging
import jk_utils

from .Spooler import Spooler
from .SpoolInfo import SpoolInfo
from .FileUncompressionGuess import FileUncompressionGuess
from .impl import TARER






class Unpacker(object):

	################################################################################################################################
	## Static Helper Methods
	################################################################################################################################

	@staticmethod
	def _uncompressGZip(
			inFilePath:str,
			outFilePath:str,
			chModValueI:int = None,
			terminationFlag:jk_utils.TerminationFlag = None,
		):
		assert inFilePath != outFilePath

		with gzip.open(inFilePath, "rb") as fin:
			if chModValueI is None:
				with open(outFilePath, "wb") as fout:
					Spooler.spoolStream(fin, fout, terminationFlag)
			else:
				fdesc = os.open(outFilePath, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, chModValueI)
				with open(fdesc, "wb") as fout:
					Spooler.spoolStream(fin, fout, terminationFlag)
	#

	@staticmethod
	def _uncompressBZip2(
			inFilePath:str,
			outFilePath:str,
			chModValueI:int = None,
			terminationFlag:jk_utils.TerminationFlag = None,
		):

		assert inFilePath != outFilePath

		with bz2.open(inFilePath, "rb") as fin:
			if chModValueI is None:
				with open(outFilePath, "wb") as fout:
					Spooler.spoolStream(fin, fout, terminationFlag)
			else:
				fdesc = os.open(outFilePath, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, chModValueI)
				with open(fdesc, "wb") as fout:
					Spooler.spoolStream(fin, fout, terminationFlag)
	#

	@staticmethod
	def _uncompressXZ(
			inFilePath:str,
			outFilePath:str,
			chModValueI:int = None,
			terminationFlag:jk_utils.TerminationFlag = None,
		):

		assert inFilePath != outFilePath

		with lzma.open(inFilePath, "rb") as fin:
			if chModValueI is None:
				with open(outFilePath, "wb") as fout:
					Spooler.spoolStream(fin, fout, terminationFlag)
			else:
				fdesc = os.open(outFilePath, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, chModValueI)
				with open(fdesc, "wb") as fout:
					Spooler.spoolStream(fin, fout, terminationFlag)
	#

	################################################################################################################################
	## Static Public Methods
	################################################################################################################################

	#
	# Untar the specified <c>tar</c> archive and write the contents to the specified directory.
	#
	# @param	str srcTarFilePath					(required) The tar archive to read
	# @param	str destDirFilePath					(required) The target directory to unpack all data to
	# @param	AbstractLogger log					(required) A logger to write log information to
	#
	@staticmethod
	def untarToDir(
			*args,
			srcTarFilePath:str,
			destDirPath:str,
			log:jk_logging.AbstractLogger,
		) -> None:

		if args:
			raise Exception("Invoke this method with named arguments only!")

		assert isinstance(srcTarFilePath, str)
		assert os.path.isfile(srcTarFilePath)
		assert isinstance(destDirPath, str)
		assert isinstance(log, jk_logging.AbstractLogger)

		# ----

		with log.descend("Unpacking " + repr(srcTarFilePath) + " ...", logLevel=jk_logging.EnumLogLevel.NOTICE) as log2:
			srcTarFilePath = os.path.abspath(srcTarFilePath)
			assert os.path.isfile(srcTarFilePath)
			destDirPath = os.path.abspath(destDirPath)

			TARER.untarToDir(
				absTarFilePath = srcTarFilePath,
				absDestDirPath = destDirPath,
				log = log2,
			)
	#

	#
	# Guess the type of uncompression and calculate the uncompression file path
	# An exception is raised on error.
	#
	@staticmethod
	def guessCompressionFromFilePathE(
			filePath:str
		) -> FileUncompressionGuess:

		assert isinstance(filePath, str)

		if filePath.endswith(".gz"):
			return FileUncompressionGuess("gz", filePath[:-3])
		elif filePath.endswith(".bz2"):
			return FileUncompressionGuess("bz2", filePath[:-4])
		elif filePath.endswith(".xz"):
			return FileUncompressionGuess("xz", filePath[:-3])
		elif filePath.endswith(".tgz"):
			return FileUncompressionGuess("gz", filePath[:-2] + "ar")
		elif filePath.endswith(".tbz2"):
			return FileUncompressionGuess("bz2", filePath[:-3] + "ar")
		elif filePath.endswith(".txz"):
			return FileUncompressionGuess("xz", filePath[:-2] + "ar")
		else:
			raise Exception("Can't guess compression!")
	#

	#
	# Guess the type of uncompression and calculate the uncompression file path
	#
	@staticmethod
	def guessCompressionFromFilePathN(
			filePath:str
		) -> typing.Union[FileUncompressionGuess,None]:

		assert isinstance(filePath, str)

		if filePath.endswith(".gz"):
			return FileUncompressionGuess("gz", filePath[:-3])
		elif filePath.endswith(".bz2"):
			return FileUncompressionGuess("bz2", filePath[:-4])
		elif filePath.endswith(".xz"):
			return FileUncompressionGuess("xz", filePath[:-3])
		elif filePath.endswith(".tgz"):
			return FileUncompressionGuess("gz", filePath[:-2] + "ar")
		elif filePath.endswith(".tbz2"):
			return FileUncompressionGuess("bz2", filePath[:-3] + "ar")
		elif filePath.endswith(".txz"):
			return FileUncompressionGuess("xz", filePath[:-2] + "ar")
		else:
			return None
	#

	#
	# Decompress the specified file.
	#
	# @param	str filePath						(required) The path of the file to uncompress.
	# @param	str toFilePath						(optional) The path of the file to write the uncompressed data to.
	# @param	str toDirPath						(optional) A directory to write the data to. If specified this overrides the
	#												regular directory the resulting file will be created in.
	# @param	bool bDeleteOriginal				(required) If <c>True</c> the source file will be deleted after successfull decompression.
	# @param	int|str|ChModValue chModValue		(optional) If specified this change-mode value will be used to set the permissions of
	#												the created file.
	# @param	TerminationFlag terminationFlag		(optional) A termination flag for graceful asynchroneous termination.
	# @param	AbstractLogger log					(required) A logger to write log information to
	# @return	str									Returns the path of the result file.
	#
	@staticmethod
	def uncompressFile(
			*args,
			filePath:str,
			toFilePath:str = None,
			toDirPath:str = None,
			bDeleteOriginal:bool = False,
			chModValue:typing.Union[int,str,jk_utils.ChModValue,None] = None,
			terminationFlag:jk_utils.TerminationFlag = None,
			log:jk_logging.AbstractLogger,
		) -> str:

		if args:
			raise Exception("Invoke this method with named arguments only!")

		assert isinstance(filePath, str)
		if toFilePath is not None:
			assert isinstance(toFilePath, str)
		if toDirPath is not None:
			assert isinstance(toDirPath, str)
			assert os.path.isdir(toDirPath)
		assert isinstance(bDeleteOriginal, bool)
		chModValue = jk_utils.ChModValue.createN(chModValue)
		chModValueI = None if chModValue is None else chModValue.toInt()
		assert isinstance(log, jk_logging.AbstractLogger)

		# ----

		with log.descend("Uncompressing " + repr(filePath) + " ...", logLevel=jk_logging.EnumLogLevel.NOTICE) as log2:
			filePath = os.path.abspath(filePath)
			assert os.path.isfile(filePath)

			guess = Unpacker.guessCompressionFromFilePathE(filePath)
			if toFilePath is None:
				toFilePath = guess.toFilePath
			if toDirPath is not None:
				# override already existing directory
				_fileName = os.path.basename(toFilePath)
				toFilePath = os.path.join(toDirPath, _fileName)

			if guess.compression in [ "gz", "gzip" ]:
				name = "gzip"
				m = Unpacker._uncompressGZip
			elif guess.compression in [ "bz2", "bzip2" ]:
				name = "bzip2"
				m = Unpacker._uncompressBZip2
			elif guess.compression in [ "xz" ]:
				name = "xz"
				m = Unpacker._uncompressXZ
			else:
				raise Exception("Unknown compression: " + repr(guess.compression))

			log.notice("Unpacking with " + name + " ...")

			orgFileSize = os.path.getsize(filePath)

			# TODO: check if target file already exists

			m(filePath, toFilePath, chModValueI, terminationFlag)

			resultFileSize = os.path.getsize(toFilePath)
			uncompressionFactor = round(100 * orgFileSize / resultFileSize, 2)
			log.notice("Uncompression factor: {}%".format(uncompressionFactor))

			if bDeleteOriginal:
				if os.path.isfile(filePath):
					os.unlink(filePath)
			else:
				if not os.path.isfile(filePath):
					raise Exception("Implementation error!")

			return toFilePath
	#

	@staticmethod
	def listTarContents(
			*args,
			tarFilePath:str,
			log:jk_logging.AbstractLogger,
		) -> typing.List[str]:

		if args:
			raise Exception("Invoke this method with named arguments only!")

		assert isinstance(tarFilePath, str)

		# ----

		tarFilePath = os.path.abspath(tarFilePath)
		assert os.path.isfile(tarFilePath)

		return TARER.listTarContents(tarFilePath, log)
	#

	@staticmethod
	def listZipContents(
			*args,
			zipFilePath:str,
			log:jk_logging.AbstractLogger,
			bIncludeDirectories:bool = True,
		) -> typing.List[str]:

		if args:
			raise Exception("Invoke this method with named arguments only!")

		assert isinstance(zipFilePath, str)

		# ----

		zipFilePath = os.path.abspath(zipFilePath)
		assert os.path.isfile(zipFilePath)

		ret = set()
		with zipfile.PyZipFile(zipFilePath, "r") as zipf:
			for relPath in zipf.namelist():
				ret.add(relPath)
				if bIncludeDirectories:
					if "\\" in relPath:
						raise Exception("????")
					if "/" in relPath:
						ret.add(os.path.dirname(relPath))

		return sorted(ret)
	#

	@staticmethod
	def listContents(
			*args,
			filePath:str,
			packing:str,
			log:jk_logging.AbstractLogger,
		) -> typing.List[str]:

		if packing == "tar":
			return Unpacker.listTarContents(
				*args,
				tarFilePath=filePath,
				log=log,
			)
		elif packing == "zip":
			return Unpacker.listZipContents(
				*args,
				zipFilePath=filePath,
				log=log,
			)
		else:
			raise Exception("Invalid packaging identifier specified: {}".format(packing))
	#

	#
	# Untar the specified <c>zip</c> archive and write the contents to the specified directory.
	#
	# @param	str srcZipFilePath					(required) The tar archive to read
	# @param	str destDirFilePath					(required) The target directory to unpack all data to
	# @param	AbstractLogger log					(required) A logger to write log information to
	#
	@staticmethod
	def unzipToDir(
			*args,
			srcZipFilePath:str,
			destDirPath:str,
			log:jk_logging.AbstractLogger,
		) -> None:

		if args:
			raise Exception("Invoke this method with named arguments only!")

		assert isinstance(srcZipFilePath, str)
		assert os.path.isfile(srcZipFilePath)
		assert isinstance(destDirPath, str)
		assert isinstance(log, jk_logging.AbstractLogger)

		# ----

		absDestDirPath = os.path.abspath(destDirPath)

		with log.descend("Unpacking " + repr(srcZipFilePath) + " ...", logLevel=jk_logging.EnumLogLevel.NOTICE) as log2:
			srcZipFilePath = os.path.abspath(srcZipFilePath)
			assert os.path.isfile(srcZipFilePath)

			with zipfile.PyZipFile(srcZipFilePath, "r") as zipf:
				return zipf.extractall(absDestDirPath)
	#

	@staticmethod
	def unpackToDir(
			*args,
			srcFilePath:str,
			destDirPath:str,
			packing:str,
			log:jk_logging.AbstractLogger,
		) -> None:

		if packing == "tar":
			Unpacker.untarToDir(
				*args,
				srcTarFilePath=srcFilePath,
				destDirPath=destDirPath,
				log=log,
			)
		elif packing == "zip":
			Unpacker.unzipToDir(
				*args,
				srcZipFilePath=srcFilePath,
				destDirPath=destDirPath,
				log=log,
			)
		else:
			raise Exception("Invalid packaging identifier specified: {}".format(packing))
	#

#

















import os
import typing
import tarfile

import jk_typing
import jk_utils
import jk_logging
import jk_json
import jk_prettyprintobj
import jk_simpleexec





class PosixExternalTar(object):

	_TAR_PATH = "/bin/tar"

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	@jk_typing.checkFunctionSignature()
	def __init__(self):
		assert os.path.isfile(PosixExternalTar._TAR_PATH)
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	#
	# Tar the contents of a directory.
	# An exception is raised on error.
	# Before an error is raised the logger may be used to provide a more detailed error message.
	#
	# @param	str absDestTarFilePath			The (absolute) output file path
	# @param	str absSrcDirPath				The (absolute) source directory that contains all files and subdirectories to include
	# @param	str[] filesAndDirsToInclude		The files and directory names to include
	# @param	AbstractLogger log				The logger to use (if there is anything to log, e.g. error information)
	#
	@jk_typing.checkFunctionSignature()
	def tarDirContents(self,
			absDestTarFilePath:str,
			absSrcDirPath:str,
			filesAndDirsToInclude:typing.List[str],
			log:jk_logging.AbstractLogger,
		) -> None:

		log.notice("Invoking " + PosixExternalTar._TAR_PATH + " ...")
		cmdResult = jk_simpleexec.invokeCmd2(
			cmdPath = PosixExternalTar._TAR_PATH,
			cmdArgs = [
					"-cf",
					absDestTarFilePath
				] + filesAndDirsToInclude,
			workingDirectory = absSrcDirPath,
		)

		if cmdResult.returnCode != 0:
			cmdResult.dump(printFunc=log.error)
			raise Exception("Failed to run 'tar'!")
	#

	@jk_typing.checkFunctionSignature()
	def listTarContents(self,
			absTarFilePath:str,
			log:jk_logging.AbstractLogger,
		) -> typing.List[str]:

		log.notice("Invoking " + PosixExternalTar._TAR_PATH + " ...")
		cmdResult = jk_simpleexec.invokeCmd2(
			cmdPath = PosixExternalTar._TAR_PATH,
			cmdArgs = [
				"-tf",
				absTarFilePath
			],
		)

		if cmdResult.returnCode != 0:
			cmdResult.dump(printFunc=log.error)
			raise Exception("Failed to run 'tar'!")

		return cmdResult.stdOutLines
	#

	@jk_typing.checkFunctionSignature()
	def untarToDir(self,
			absTarFilePath:str,
			absDestDirPath:str,
			log:jk_logging.AbstractLogger,
		) -> None:

		if not os.path.isdir(absDestDirPath):
			os.makedirs(absDestDirPath)

		log.notice("Invoking " + PosixExternalTar._TAR_PATH + " ...")
		cmdResult = jk_simpleexec.invokeCmd2(
			cmdPath = PosixExternalTar._TAR_PATH,
			cmdArgs = [
					"-xf",
					absTarFilePath,
					"-C",
					absDestDirPath
				],
			workingDirectory = absDestDirPath,
		)

		if cmdResult.returnCode != 0:
			cmdResult.dump(printFunc=log.error)
			raise Exception("Failed to run 'tar'!")
	#

#










import os
import typing
import tarfile
import gzip
import bz2
import lzma
import time
import re
import zipfile

import jk_simpleexec
import jk_logging
import jk_utils
import jk_dirwalker

from .Spooler import Spooler
from .SpoolInfo import SpoolInfo
from .impl import TARER
from ._NameMatcher import _NameMatcher





class Packer(object):

	################################################################################################################################
	## Static Helper Methods
	################################################################################################################################

	@staticmethod
	def _compressGZip(
			inFilePath:str,
			outFilePath:str,
			chModValueI:int = None,
			terminationFlag:jk_utils.TerminationFlag = None
		):

		assert inFilePath != outFilePath

		with open(inFilePath, "rb") as fin:
			if chModValueI is None:
				with gzip.open(outFilePath, "wb") as fout:
					Spooler.spoolStream(fin, fout, terminationFlag)
			else:
				fdesc = os.open(outFilePath, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, chModValueI)
				with open(fdesc, "wb") as fout2:
					with gzip.open(fout2, "wb") as fout:
						Spooler.spoolStream(fin, fout, terminationFlag)
	#

	@staticmethod
	def _compressBZip2(
			inFilePath:str,
			outFilePath:str,
			chModValueI:int = None,
			terminationFlag:jk_utils.TerminationFlag = None
		):

		assert inFilePath != outFilePath

		with open(inFilePath, "rb") as fin:
			if chModValueI is None:
				with bz2.open(outFilePath, "wb") as fout:
					Spooler.spoolStream(fin, fout, terminationFlag)
			else:
				fdesc = os.open(outFilePath, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, chModValueI)
				with open(fdesc, "wb") as fout2:
					with bz2.open(fout2, "wb") as fout:
						Spooler.spoolStream(fin, fout, terminationFlag)
	#

	@staticmethod
	def _compressXZ(
			inFilePath:str,
			outFilePath:str,
			chModValueI:int = None,
			terminationFlag:jk_utils.TerminationFlag = None
		):

		assert inFilePath != outFilePath

		with open(inFilePath, "rb") as fin:
			if chModValueI is None:
				with lzma.open(outFilePath, "wb") as fout:
					Spooler.spoolStream(fin, fout, terminationFlag)
			else:
				fdesc = os.open(outFilePath, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, chModValueI)
				with open(fdesc, "wb") as fout2:
					with lzma.open(fout2, "wb") as fout:
						Spooler.spoolStream(fin, fout, terminationFlag)
	#

	#
	# @return		str name
	# @return		str ext
	# @return		callable m
	#
	@staticmethod
	def _getCompressionParams(compression:str):
		assert isinstance(compression, str)

		if compression in [ "gz", "gzip" ]:
			return "gzip", ".gz", Packer._compressGZip
		elif compression in [ "bz2", "bzip2" ]:
			return "bzip2", ".bz2", Packer._compressBZip2
		elif compression in [ "xz" ]:
			return "xz", ".xz", Packer._compressXZ
		else:
			raise Exception("Unknown compression: " + repr(compression))
	#

	################################################################################################################################
	## Static Public Methods
	################################################################################################################################

	#
	# This method selects files and directories (non-recursively) from the specified source directory.
	#
	# This is a helper method only. It is used by other method(s) such as <c>tarDirContents()</c> for filtering.
	# However, it is quite useful for other purposes not related to compression. This is the reason why it is made
	# public in this module though it is not related to compression in a strict sense.
	#
	@staticmethod
	def selectFilesAndDirs(
			dirPath:str,
			filesAndDirsToInclude:typing.Iterable[typing.Union[str,re.Match]] = None,
			filesAndDirsToExclude:typing.Iterable[typing.Union[str,re.Match]] = None,
			bDebugIncludeExclude:bool = False,
			log:jk_logging.AbstractLogger = None,
		) -> typing.List[str]:

		if bDebugIncludeExclude:
			log.notice(":: filesAndDirsToInclude = " + str(filesAndDirsToInclude))
			log.notice(":: filesAndDirsToExclude = " + str(filesAndDirsToExclude))

		_includeMatcher = _NameMatcher(filesAndDirsToInclude, bDefaultIfNone=True)
		_excludeMatcher = _NameMatcher(filesAndDirsToExclude, bDefaultIfNone=False)

		entries = []
		if bDebugIncludeExclude:
			for fe in os.scandir(dirPath):
				if _includeMatcher.matchDebug(fe.name, "include candidate", "exclude", log) \
						and not _excludeMatcher.matchDebug(fe.name, "exclude", "don't exclude", log):
					log.notice("-> including: " + fe.name)
					entries.append(fe.name)
		else:
			for fe in os.scandir(dirPath):
				if _includeMatcher.match(fe.name) and not _excludeMatcher.match(fe.name):
					entries.append(fe.name)

		return sorted(entries)
	#

	#
	# Pack the contents of the specified directory in a tar file.
	#
	# @param	str srcDirPath						(required) The directory to pack
	# @param	str destTarFilePath					(required) The tar file to create
	# @param	str[] filesAndDirsToInclude			(optional) The file and directorie names (without path!) to include.
	#												If <c>None</c> is specified the source directory is scanned and all
	#												files and directories found there will be included automatically.
	# @param	AbstractLogger log					(required) A logger to write log information to
	# @param	Callable[TarInfo]:TarInfo filter	(optional) An optional filter function that is directly passed on to the
	#												python tarfile API. This filter function receives a TarInfo record of
	#												the current file to pack and can either return the same or a modified
	#												version of the TarInfo record or None to indicate that
	#												this entry should be excluded.
	#
	@staticmethod
	def tarDirContents(
			*args,
			srcDirPath:str,
			destTarFilePath:str,
			chModValue:typing.Union[int,str,jk_utils.ChModValue,None] = None,
			filesAndDirsToInclude:typing.Iterable[typing.Union[str,re.Match]] = None,
			filesAndDirsToExclude:typing.Iterable[typing.Union[str,re.Match]] = None,
			filter:typing.Callable[[tarfile.TarInfo],typing.Union[tarfile.TarInfo,None]] = None,
			log:jk_logging.AbstractLogger,
			bDebugIncludeExclude:bool = False,
		) -> typing.Union[str,None]:

		if args:
			raise Exception("Invoke this method with named arguments only!")

		assert isinstance(srcDirPath, str)
		assert isinstance(destTarFilePath, str)
		chModValue = jk_utils.ChModValue.createN(chModValue)
		chModValueI = None if chModValue is None else chModValue.toInt()
		assert isinstance(log, jk_logging.AbstractLogger)
		assert isinstance(bDebugIncludeExclude, bool)

		# ----------------------------------------------------------------

		with log.descend("Tar-Packing " + repr(srcDirPath) + " ...", logLevel=jk_logging.EnumLogLevel.NOTICE) as log2:
			srcDirPath = os.path.abspath(srcDirPath)
			assert os.path.isdir(srcDirPath)
			destTarFilePath = os.path.abspath(destTarFilePath)

			# --------
			# create a flat list of file and directory names that are to include from srcDirPath

			_selectedEntries = Packer.selectFilesAndDirs(srcDirPath, filesAndDirsToInclude, filesAndDirsToExclude, bDebugIncludeExclude, log2)
			if not _selectedEntries:
				log2.notice("No files and directories to include.")
				return None

			# --------
			# do the packing

			_oldmask = os.umask(0o777 ^ chModValueI) if chModValueI is not None else None
			try:
				TARER.tarDirContents(
					destTarFilePath,
					srcDirPath,
					_selectedEntries,
					log2,
					filter=filter,
				)
			finally:
				if _oldmask is not None:
					os.umask(_oldmask)
					if os.path.isfile(destTarFilePath):
						os.chmod(destTarFilePath, chModValueI)	# required as tar will not set the execute bit

		# ----------------------------------------------------------------

		return destTarFilePath
	#

	#
	# Pack the contents of the specified directory in a zip file.
	#
	# @param	str srcDirPath						(required) The directory to pack
	# @param	str destZipFilePath					(required) The tar file to create
	# @param	str[] filesAndDirsToInclude			(optional) The file and directorie names (without path!) to include.
	#												If <c>None</c> is specified the source directory is scanned and all
	#												files and directories found there will be included automatically.
	# @param	AbstractLogger log					(required) A logger to write log information to
	#
	@staticmethod
	def zipDirContents(
			*args,
			srcDirPath:str,
			destZipFilePath:str,
			chModValue:typing.Union[int,str,jk_utils.ChModValue,None] = None,
			filesAndDirsToInclude:typing.Iterable[typing.Union[str,re.Match]] = None,
			filesAndDirsToExclude:typing.Iterable[typing.Union[str,re.Match]] = None,
			log:jk_logging.AbstractLogger,
			bDebugIncludeExclude:bool = False,
		) -> typing.Union[str,None]:

		if args:
			raise Exception("Invoke this method with named arguments only!")

		assert isinstance(srcDirPath, str)
		assert isinstance(destZipFilePath, str)
		chModValue = jk_utils.ChModValue.createN(chModValue)
		chModValueI = None if chModValue is None else chModValue.toInt()
		assert isinstance(log, jk_logging.AbstractLogger)
		assert isinstance(bDebugIncludeExclude, bool)

		# ----------------------------------------------------------------

		with log.descend("Zip-Packing " + repr(srcDirPath) + " ...", logLevel=jk_logging.EnumLogLevel.NOTICE) as log2:
			srcDirPath = os.path.abspath(srcDirPath)
			assert os.path.isdir(srcDirPath)
			destZipFilePath = os.path.abspath(destZipFilePath)

			# --------
			# create a flat list of file and directory names that are to include from srcDirPath

			_selectedEntries = Packer.selectFilesAndDirs(srcDirPath, filesAndDirsToInclude, filesAndDirsToExclude, bDebugIncludeExclude, log2)
			if not _selectedEntries:
				log2.notice("No files and directories to include.")
				return None

			# --------
			# unfortunately the zipfile module does no recursive descend on its own. therefore we create a list of files to include into the
			# ZIP file manually.

			filesToIncludeRelPaths = []
			w = jk_dirwalker.DirWalker(emitFilter = jk_dirwalker.StdEmitFilter.newFromDisabled(
					emitRegularFiles = True,
				),
				raiseErrors = True,
			)
			for _selectedEntry in _selectedEntries:
				_absPath = os.path.join(srcDirPath, _selectedEntry)
				if os.path.isfile(_absPath):
					# files
					filesToIncludeRelPaths.append(_selectedEntry)
				elif os.path.isdir(_absPath):
					# directories
					for x in w.scandir(_absPath):
						filesToIncludeRelPaths.append(os.path.join(_selectedEntry, x.relFilePath))
				else:
					# ignore
					pass

			# --------
			# do the packing

			_oldmask = os.umask(0o777 ^ chModValueI) if chModValueI is not None else None
			try:
				with zipfile.PyZipFile(destZipFilePath, "w") as zipf:
					for selectedEntry in filesToIncludeRelPaths:
						assert isinstance(selectedEntry, str)
						localFilePath = os.path.join(srcDirPath, selectedEntry.replace("/", os.path.sep))
						log2.debug("Writing: " + localFilePath + " as " + selectedEntry)
						zipf.write(localFilePath, selectedEntry)
			finally:
				if _oldmask is not None:
					os.umask(_oldmask)
					if os.path.isfile(destZipFilePath):
						os.chmod(destZipFilePath, chModValueI)	# required as tar will not set the execute bit

		# ----------------------------------------------------------------

		return destZipFilePath
	#

	@staticmethod
	def isValidCompression(compression:str) -> bool:
		try:
			Packer._getCompressionParams(compression)
			return True
		except:
			return False
	#

	@staticmethod
	def isAlreadyCompressed(filePath:str) -> bool:
		fileName = os.path.basename(filePath)
		_, fileExt = os.path.splitext(fileName)
		if not fileExt:
			return None

		assert fileExt.startswith(".")
		fileExt = fileExt[1:]

		try:
			Packer._getCompressionParams(fileExt)
			return True
		except:
			return False
	#

	#
	# Pack the contents of the specified directory.
	#
	# @param	str srcDirPath						(required) The directory to pack
	# @param	str destTarFilePath					(required) The file to create
	# @param	str packaging						(required) Either specify "tar" or "zip" here.
	# @param	str[] filesAndDirsToInclude			(optional) The file and directorie names (without path!) to include.
	#												If <c>None</c> is specified the source directory is scanned and all
	#												files and directories found there will be included automatically.
	# @param	AbstractLogger log					(required) A logger to write log information to
	#
	@staticmethod
	def packDirContents(
			*args,
			srcDirPath:str,
			destTarFilePath:str,
			packing:str,
			chModValue:typing.Union[int,str,jk_utils.ChModValue,None] = None,
			filesAndDirsToInclude:typing.Iterable[typing.Union[str,re.Match]] = None,
			filesAndDirsToExclude:typing.Iterable[typing.Union[str,re.Match]] = None,
			log:jk_logging.AbstractLogger,
			bDebugIncludeExclude:bool = False,
		) -> typing.Union[str,None]:

		if packing == "tar":
			return Packer.tarDirContents(
				*args,
				srcDirPath=srcDirPath,
				destTarFilePath=destTarFilePath,
				chModValue=chModValue,
				filesAndDirsToInclude=filesAndDirsToInclude,
				filesAndDirsToExclude=filesAndDirsToExclude,
				log=log,
				bDebugIncludeExclude=bDebugIncludeExclude,
			)
		elif packing == "zip":
			return Packer.zipDirContents(
				*args,
				srcDirPath=srcDirPath,
				destZipFilePath=destTarFilePath,
				chModValue=chModValue,
				filesAndDirsToInclude=filesAndDirsToInclude,
				filesAndDirsToExclude=filesAndDirsToExclude,
				log=log,
				bDebugIncludeExclude=bDebugIncludeExclude,
			)
		else:
			raise Exception("Invalid packaging identifier specified: {}".format(packing))
	#


	#
	# Compress the specified file.
	#
	# @param	str filePath						(required) The path of the file to compress.
	# @param	str toFilePath						(optional) The path of the file to write the compressed data to.
	#												If <c>None</c> a new file path is created with suitable exension based on the
	#												compression type.
	# @param	str toDirPath						(optional) A directory to write the data to. If specified this overrides the
	#												regular directory the resulting file will be created in.
	# @param	str compression						(required) The compression. Valid values are: "gz", "gzip", "bz2", "bzip2", "xz"
	# @param	bool bDeleteOriginal				(required) If <c>True</c> the source file will be deleted after successfull compression.
	# @param	int|str|ChModValue chModValue		(optional) If specified this change-mode value will be used to set the permissions of
	#												the created file.
	# @param	TerminationFlag terminationFlag		(optional) A termination flag for graceful asynchroneous termination.
	# @param	AbstractLogger log					(required) A logger to write log information to
	# @return	str									Returns the path of the result file.
	#
	@staticmethod
	def compressFile(
			*args,
			filePath:str,
			toFilePath:str = None,
			toDirPath:str = None,
			compression:str,
			bDeleteOriginal:bool = False,
			chModValue:typing.Union[int,str,jk_utils.ChModValue,None] = None,
			terminationFlag:jk_utils.TerminationFlag = None,
			log:jk_logging.AbstractLogger,
		) -> SpoolInfo:

		if args:
			raise Exception("Invoke this method with named arguments only!")

		assert isinstance(filePath, str)
		if toFilePath is not None:
			assert isinstance(toFilePath, str)
		if toDirPath is not None:
			assert isinstance(toDirPath, str)
			assert os.path.isdir(toDirPath)
		assert isinstance(compression, str)
		assert isinstance(bDeleteOriginal, bool)
		chModValue = jk_utils.ChModValue.createN(chModValue)
		chModValueI = None if chModValue is None else chModValue.toInt()
		assert isinstance(log, jk_logging.AbstractLogger)

		# ----

		with log.descend("Compressing " + repr(filePath) + " ...", logLevel=jk_logging.EnumLogLevel.NOTICE) as log2:
			filePath = os.path.abspath(filePath)
			assert os.path.isfile(filePath)

			compressionName, compressionFileExt, m = Packer._getCompressionParams(compression)

			log.notice("Packing with " + compressionName + " ...")

			orgFileSize = os.path.getsize(filePath)

			if toFilePath is None:
				toFilePath = filePath + compressionFileExt
			if toDirPath is not None:
				# override already existing directory
				_fileName = os.path.basename(toFilePath)
				toFilePath = os.path.join(toDirPath, _fileName)

			# TODO: check if target file already exists

			tStart = time.time()
			m(filePath, toFilePath, chModValueI, terminationFlag)
			tDuration = time.time() - tStart

			resultFileSize = os.path.getsize(toFilePath)

			if bDeleteOriginal:
				if os.path.isfile(filePath):
					os.unlink(filePath)
			else:
				if not os.path.isfile(filePath):
					raise Exception("Implementation error!")

			return SpoolInfo(filePath, toFilePath, compressionName, compressionFileExt, orgFileSize, resultFileSize, tDuration)
	#

#




























import os


#if os.name == "posix":
#	from .PosixExternalTar import PosixExternalTar
#	TARER = PosixExternalTar()
#else:
#	from .PythonNativeTar import PythonNativeTar
#	TARER = PythonNativeTar()

from .PythonNativeTar import PythonNativeTar
TARER = PythonNativeTar()